/**
 * \author Thomas Buechler
 * @brief defintion of a stack that grows linear
 */

#ifndef EDSM_MYSTACK_DEF
#define EDSM_MYSTACK_DEF


/**
 * @brief stack data structure that uses a sdsl::int_vector
 * 
 * grows linear if resize_value > 0 is given (if vector is full resize_value empty spots will be added)
 * grows exponentially if resize_value = 0 (if vector is full the data vector will be doubled) <- DEFAULT
 */
template<uint8_t t_width>
struct my_stack{
private:
 	static const uint32_t DEFAULT_RESIZE_VALUE = 0;
 	static const uint32_t DEFAULT_SIZE = 10;
	
public:	
	uint32_t resize_value;	
	uint32_t top_pointer;
	sdsl::int_vector<t_width> data_vector;
	sdsl::int_vector<t_width> neg_one;	
	
	

	my_stack(uint32_t size = my_stack::DEFAULT_SIZE, uint32_t resize_value = my_stack::DEFAULT_RESIZE_VALUE);
	my_stack(const my_stack &ms);
	
	
	void push(uint64_t val);	
	uint64_t top()const;
	void pop();
	
	void remove(uint32_t i); //removes ith element in stack, fills empty space with top and deletes top;
	
	
	uint32_t size() const ;
	
	void print (std::ostream& outputStream) const;	
	bool empty() const ;
	
	void resize();
	
	uint64_t operator[] (uint32_t i) const;
};

template <uint8_t t_width> std::ostream& operator <<(std::ostream& outputStream, const my_stack<t_width>& stack);


/** @brief empty constructor*/
template<uint8_t t_width>
my_stack<t_width>::my_stack(uint32_t defaul_size, uint32_t resize_value):resize_value(resize_value){
	top_pointer = 0;
	data_vector = sdsl::int_vector<t_width>(defaul_size,0,t_width);
	neg_one	= sdsl::int_vector<t_width>(1,-1,t_width);
}

/** @brief copy constructor */
template<uint8_t t_width>
my_stack<t_width>::my_stack(const my_stack &ms){
	top_pointer = ms.top_pointer;
	resize_value = ms.resize_value;
	data_vector = sdsl::int_vector<t_width>(ms.data_vector);
	neg_one	= sdsl::int_vector<t_width>(1,-1,t_width);
};	
	

template<uint8_t t_width>
void my_stack<t_width>::push(uint64_t val){
	if(top_pointer == data_vector.size()){
		if(resize_value == 0)
			data_vector.resize((data_vector.size()+1)*2);
		else
			data_vector.resize(data_vector.size()+resize_value);
	}
	data_vector[top_pointer] = val;		
	top_pointer++; 
	if(top_pointer==0)
		throw std::runtime_error("my_stack: overflow");		
};

template<uint8_t t_width>
uint64_t my_stack<t_width>::top()const{
	if(top_pointer == 0)
		throw std::out_of_range("my_stack: stack is empty");
	return data_vector[top_pointer-1];
};

template<uint8_t t_width>
void my_stack<t_width>::pop(){if(top_pointer > 0)top_pointer--;};

/** @brief print for debugging*/
template<uint8_t t_width>
void my_stack<t_width>::print (std::ostream& os) const { 
	os << "MY_STACK: ";
	for( uint16_t i = top_pointer - 1; i != (uint16_t) -1; i--){
		os << data_vector[i] << " ";
	}
	os << "#";
};
	
/** @brief replaces ith element in stack with top and removes top */
template<uint8_t t_width>
void my_stack<t_width>::remove(uint32_t i){
	if(i >= top_pointer)
		throw std::invalid_argument("my_stack: index greater then stack");
	
	data_vector[i]=data_vector[top_pointer-1];
	pop();		
};
	
template<uint8_t t_width>
bool my_stack<t_width>::empty() const {return top_pointer == 0;};

template<uint8_t t_width>
void my_stack<t_width>::resize(){data_vector.resize(top_pointer);}

template<uint8_t t_width>
uint32_t my_stack<t_width>::size() const {return top_pointer;}

template<uint8_t t_width>
uint64_t my_stack<t_width>::operator[] (uint32_t i) const {return data_vector[i];}



template <uint8_t t_width>
std::ostream& operator <<(std::ostream& os, const my_stack<t_width>& stack){
	stack.print(os);
	return os;
}

#endif
