#include "my_stack.cpp"
#include "my_int_vector_buffer.cpp"
#include "kmer.cpp"

#include <algorithm>    // std::sort
#include <utility>      // std::pair, std::make_pair
#include <tuple>        // std::tuple, std::get, std::tie, std::ignore

#include "rsort.cpp"

using namespace std;
using namespace sdsl;


template <uint8_t t_width, typename int_type>
struct fasta_read{
	typedef tuple<int_type,int_type,int_type> int_type_triple;
	typedef tuple<string,uint32_t,int_type,bool> alignment_type;

	string	id;			//name of seq
	string	sequence;		//
	string	qual;

	//[EOC]QUERY KMERS
	//first  = id of kmer
	//second = start of this kmer in pattern
	vector<pair<int_type,int_type>> kmer_pairs;
	bool filled_1;

	//QUERY POSITIONS
	//first  = position in EDS
	//second = position in read
	vector<pair<int_type,int_type>> pos_pairs;
	bool filled_3;

	//QUERY HOTSPOTS
	vector< int_type_triple > hotspots;
	//0 positions in EDS that are in a hot spot
	//1 coresponding positions in pattern
	//2 number of seeds at this hotspot
	bool filled_4;

	//ALIGNMENTS
	vector<alignment_type> alignments;
	bool filled_5;

	fasta_read(){
		filled_1 = filled_3 = filled_4 = filled_5 = false;
	}
	
	fasta_read(string id ,string sequence, string qual):id(id),sequence(sequence), qual(qual){
		filled_1 = filled_3  = filled_4 = filled_5 = false;
	}

	fasta_read(ifstream & fastq_s){
		using namespace std;
		getline(fastq_s, id);
		if(!fastq_s.good()) throw runtime_error ("NO READ IN STREAM");
		getline(fastq_s, sequence);
		getline(fastq_s, qual);  //plus
		getline(fastq_s, qual); //qual
		if(!fastq_s.good()){
			edsm_io::print_error("READ " + id + " incomplete");
			throw runtime_error ("NO READ IN STREAM");
		}
	}

	bool operator==(const fasta_read<t_width,int_type> & fr){
		return id == fr.id;
	}

	void delete_kmer_vectors(){
		vector<pair<int_type,int_type>>(0).swap(kmer_pairs);
		filled_1 = false;
	}

	void delete_pos_vectors(){
		vector<pair<int_type,int_type>>(0).swap(pos_pairs);
		filled_3 = false;
	}

	void delete_hints(){
		vector< tuple<int_type,int_type,int_type> >(0).swap(hotspots);
		filled_4 = false;
	}

	void delete_alignment(){
		vector<alignment_type>(0).swap(alignments);
		filled_5 = false;
	};

	/**
	 * @brief gets all seeds
	 *
	void get_seed_kmers(const uint32_t shift, uint32_t k){
		assert(k <= sequence.length());

		uint32_t seed_count = (sequence.length()-k) / shift+1;

		kmer_pairs	= vector<pair<int_type,int_type>>(seed_count);

		uint32_t i = 0;
		for(	int_type  seed_begin =  0;
			seed_begin 	+  k <= sequence.size();
			seed_begin 	+= shift)
		{
			kmer_pairs[i++] = make_pair<int_type,int_type>((int_type) KMER<int_type>::to_id(sequence.substr(seed_begin, k)),(int_type) seed_begin);
		}
		filled_1 = true;
	};*/
	
	/**
	 * @brief generate fragments from sequence
	 */
	void get_fragments(const uint32_t fragment_count, const uint32_t k){
		assert(k <= sequence.length());
		
		/*if( k*fragment_count >=  sequence.length()){
			kmer_pairs	= vector<pair<int_type,int_type>>(fragment_count);
			uint32_t fragment_begin = 0;
			for(uint32_t i = 0; i < fragment_count; i++){
				kmer_pairs[i] = make_pair<int_type,int_type>((int_type) KMER<int_type>::to_id(sequence.substr(fragment_begin, k)),(int_type) fragment_begin);
				fragment_begin += k;
			}
			return;
		}*/
		
		uint32_t this_fragment_count = fragment_count;
		
		if(  (sequence.length() - k)/ (double) fragment_count   < 1){
			edsm_io::print_error("Not possible to generate that many fragments of the sequence. Set fragment_count value to (read.length - k)");
			this_fragment_count = (sequence.length() - k);
		}
		
		kmer_pairs	= vector<pair<int_type,int_type>>(this_fragment_count);
		
		for(uint32_t i = 0; i < this_fragment_count; i++){			
			uint32_t fragment_begin = i* ((sequence.length() - k)/ (double) this_fragment_count);
			kmer_pairs[i] = make_pair<int_type,int_type>((int_type) KMER<int_type>::to_id(sequence.substr(fragment_begin, k)),(int_type) fragment_begin);
		}
		
		filled_1 = true;
	};
	
	/**
	 * @brief generate fragments from sequence
	 */
	void get_fragments(const uint32_t fragment_count, const uint32_t k, sdsl::bit_vector & indicator){
		assert(k <= sequence.length());
		
		
		uint32_t this_fragment_count = 0;
		kmer_pairs	= vector<pair<int_type,int_type>>(fragment_count);
		
		for(uint32_t i = 0; i < k && this_fragment_count < fragment_count; i++){
			for(uint32_t fragment_begin = i; fragment_begin < (sequence.length() - k) && this_fragment_count < fragment_count; fragment_begin += k){
				int_type kmer_id = KMER<int_type>::to_id(sequence.substr(fragment_begin, k));
				if( indicator[kmer_id] ){
					kmer_pairs[this_fragment_count++] = make_pair<int_type,int_type>( (int_type) kmer_id , (int_type) fragment_begin);
				}
			}
		}
		
		if( this_fragment_count < fragment_count){
			kmer_pairs.resize(this_fragment_count);
// 			edsm_io::print_error("Only generated " + to_string(this_fragment_count) + " fragment for read " + id);
		}
		
		filled_1 = true;
	};

	/**
	 * @brief queries the fragment positions
	 * @param eoc the EOC-index
	 */
	void get_positions(linear_eoc_type & eoc){
		vector<uint32_t> boundaries(kmer_pairs.size()+1);
		uint32_t b_i = 0;
		boundaries[b_i++] = 0;
		for(auto it = kmer_pairs.begin(); it != kmer_pairs.end(); it++){
			int_type k = get<0>(*it);
			uint32_t begin = eoc.table[k];
			uint32_t count = eoc.table[k+1] - begin;
			if(count > 0){
				uint64_t old_size = pos_pairs.size();
				pos_pairs.resize(old_size + count);

				//COPY VALUES FROM KMI
				for( uint32_t j  = 0; j < count; j++) pos_pairs[old_size + j] = make_pair(eoc.occurences[begin + j], get<1>(*it)) ;

				boundaries[b_i++] = pos_pairs.size();
			}
		}
		boundaries.resize(b_i);
		multi_merge(pos_pairs.begin(),boundaries);
		delete_kmer_vectors();
		filled_3 = true;
	};
	/**
	 * @brief queries the seed positions in the batch
	 * @param eoc 		the EOC vector, eoc[1] is seperator, that seperates kmers
	 * @param batch	vector containing the patterns
	 */
	//template<class eoc_type>
	//static void get_positions_in_batch_eoc(vector<fasta_read<t_width,int_type>> & batch, eoc_type & eoc);


	/**
	* @brief performs hotspot finding
	* 
	* @param window_size
	* @param window_hits 
	* @param check_col check for colliniarity
	*/
	void find_hotspots(const uint32_t window_size, const uint32_t window_hits, bool check_col);


	/**
	* @brief aligns hotspots to EDS
	* 
	* @param EDS	pangenome
	* @param D	max Dist 
	* @param max_a_c maximum number of completed alignments
	* @param max_a_t maximum number of started alignments
	* @param max_a_c maximum number of alignments in output
	* 
	*/
	void start_aligner(const string & EDS, const adjacency & adj, uint32_t D, uint32_t max_a_c, uint32_t max_a_t, uint32_t max_a_o){
		fasta_read<t_width,int_type> DUMMY_FR;
		start_aligner(EDS,adj,D,max_a_c,max_a_t,max_a_o,DUMMY_FR);
	};

	/**
	* @brief aligns hotspots to EDS with considering a reverse complement
	* 
	* @param EDS	pangenome
	* @param D	max Dist 
	* @param max_a_c maximum number of completed alignments
	* @param max_a_t maximum number of started alignments
	* @param max_a_c maximum number of alignments in output
	* @param rev_compl	reverse complement of read
	* 
	*/
	void start_aligner(const string & EDS, const adjacency & adj, uint32_t D, uint32_t max_a_c, uint32_t max_a_t, uint32_t max_a_o, fasta_read<t_width,int_type> & rev_compl);	

	/**
	* @brief writes alignment in file 
	* @param ofs	ofstream
	* @param write_failure	if true a line with name and stars in written if read could not be aligned
	* @param transform		transforms position in EDS to position in FA
	* @return number of alignments in output
	*/
	uint32_t write_alignment(ofstream & ofs, bool write_failure, const pos_EDS_to_FA_type & transform){
		using namespace std;
		assert(filled_5);
		
		uint32_t res = alignments.size();
		
		
		string 	QNAME	= id.substr(1);
		uint32_t 	FLAG		= 0;
		string	RNAME	= "*";
		uint32_t 	POS		= 0;
		uint32_t	MAPQ 	= 255; //TODO
		string 	CIGAR	= "*";
		string 	RNEXT 	= "*";
		uint32_t 	PNEXT 	= 0;
		uint32_t 	TLEN		= 0;
		string	SEQ		= sequence;
		string 	QUAL		= qual;
		
		//FOR FLAG
		uint32_t UNMAPPED = 4;
		uint32_t REV_COMP	= 16;
		uint32_t NOT_PRIM	= 256;
		
		for(uint32_t i = 0; i < alignments.size(); i++){
			RNAME 	= "*";
			POS 		= get<2>(alignments[i])+1;
			CIGAR 	= get<0>(alignments[i]);
			SEQ		= sequence;
			QUAL		= qual;
			
			FLAG = 0;
			if( i > 0 )	
				FLAG += NOT_PRIM;
			
			if(get<3>(alignments[i])){
			 	FLAG += REV_COMP;
				SEQ = edsm_encode::rev_complement(SEQ);
				QUAL = edsm_encode::rev_complement(QUAL,false);
			}
			uint32_t off = 0;
			if(!transform.empty()) tie(RNAME,POS,off) = transform(POS);
			
				
			ofs 	<< QNAME << '\t' << FLAG << '\t' << RNAME << '\t' << POS << '\t' << MAPQ << '\t' << CIGAR << '\t' << RNEXT << '\t' << PNEXT << '\t' << TLEN << '\t' << SEQ << '\t' << QUAL << '\t'
				<< "NM:i:" << ((uint32_t)get<1>(alignments[i])) <<  (off?(" XO:i:" + to_string(off)):"") << endl;
		}
		
		if(!alignments.size() && write_failure)
			ofs 	<< QNAME << '\t' << UNMAPPED << '\t' << RNAME << '\t' << POS << '\t' << MAPQ << '\t' << CIGAR << '\t' << RNEXT << '\t' << PNEXT << '\t' << TLEN << '\t' << SEQ << '\t' << QUAL <<  endl;
		delete_alignment();
		return res;
	};
};

/**
template <uint8_t t_width, class int_type>
template<class eoc_type>
void fasta_read<t_width,int_type>::get_positions_in_batch_eoc(vector<fasta_read<t_width,int_type>> & batch, eoc_type & eoc){
	sys_timer batch_time;
	batch_time.start();
	//calc numer of intervals
	uint32_t kmer_count = 0;
	for(auto r : batch){
		assert(r.filled_1);
		kmer_count += r.kmer_pairs.size();
	}

	vector<int_type_triple> all_kmers = vector<int_type_triple>(kmer_count);

	uint32_t k = 0;
	for(uint32_t i = 0; i < batch.size(); i++){
		for(uint32_t j = 0; j < batch[i].kmer_pairs.size(); j++){
			all_kmers[k++] =(  make_tuple<>(batch[i].kmer_pairs[j].first, batch[i].kmer_pairs[j].second , i )   );
		}
		batch[i].delete_kmer_vectors();
	}

	for(size_t i = 0; i < batch.size(); i++){
		batch[i].pos_pairs.reserve(edsm_align::DEFAULT_POSPAIR_CAPACITY);
	}
	
	sort(all_kmers.begin(),
		all_kmers.end(), 
		[] ( int_type_triple const& t1, int_type_triple const& t2) -> bool{ return get<0>(t1) < get<0>(t2); } //SORT BY HITS ID
	);

	
	int_type sep = eoc[1];
	size_t eoc_p = 1;
	int_type eoc_id = 0;
	int_type i = 0;
	while(true){
		
		eoc_p++;		
		int_type p = eoc[eoc_p];
		while(p == sep && eoc_p < eoc.size()){
			eoc_id++;
			eoc_p++;
			p = eoc[eoc_p];
		}
		if(eoc_p == eoc.size()) break;			
			
		
		for(; get<0>(all_kmers[i]) < eoc_id && i < all_kmers.size() ; i++);
		
		for(size_t j = i; get<0>(all_kmers[j]) == eoc_id; j++)
			batch[ get<2>(all_kmers[j]) ].pos_pairs.push_back( make_pair(p,get<1>(all_kmers[j]) ));
	}
}*/

template <uint8_t t_width, class int_type>
void fasta_read<t_width,int_type>::find_hotspots(const uint32_t window_size, const uint32_t window_hits, bool check_col){
 	assert(filled_3);//positions must be availablee in vector
	
	//sort positions by index positions
	//std::sort(pos_pairs.begin(),pos_pairs.end());
	//q_sort(pos_pairs.begin(),pos_pairs.end());

	//boundaries of search window
	uint32_t l = 0;
	uint32_t r = 0;
	
	while( l < pos_pairs.size() && r < pos_pairs.size()) {
		//shift r as far to the right as possible, i.e. [pos_i[l],pos_i[r]] < window_size
		while(r + 1 < pos_pairs.size() && pos_pairs[r+1].first < pos_pairs[l].first + window_size) 
			r++;
		
		//if interval is big enough
		if(r-l+1 >= window_hits) {
			if(check_col){
				//cl_v[i] := number of smaller positions in pos_s[l..l+i]
				std::vector<uint32_t> cl_v(r-l+1,0);								
				for(uint32_t i = 1; i < cl_v.size(); i++){
					for(uint32_t j = 0; j < i; j++)
						if(pos_pairs[l+j].second < pos_pairs[l+i].second && cl_v[i] <= cl_v[j] )
							cl_v[i] = cl_v[j] + 1;
					if(cl_v[i] + 1 >= window_hits)
						hotspots.push_back( make_tuple<>(pos_pairs[l+i].first, pos_pairs[l+i].second, cl_v[i]+1) );
				}
				
			}else
				for(uint32_t i = l; i < r; i++)
					hotspots.push_back( make_tuple<>(pos_pairs[i].first, pos_pairs[i].second, r-l+1) );
		}

		//shift r by one
		r++;
		if(r >= pos_pairs.size())
			break;

		//shift l so that [pos_i[l],pos_i[r]] < window_size holds
 		while(pos_pairs[l].first + window_size  < pos_pairs[r].first) 
 			l++;	
	}
	delete_pos_vectors();

	

	
	if(hotspots.size() > 0){ //remove duplettes TODO MAYBE DO NOT DO THAT
		
		//rsort<true>(hotspots.data(), hotspots.data() + hotspots.size(),[] (const auto& v) { return get<1>(v); });
		sort(hotspots.begin(), hotspots.end(), 
		     [] ( int_type_triple const& t1, int_type_triple const& t2) -> bool{ return get<0>(t1) < get<0>(t2); });
		
		
		uint32_t j = 0;
		for(uint32_t i = 1; i < hotspots.size(); i++){
			if( get<0>(hotspots[i]) != get<0>(hotspots[j]) || get<1>(hotspots[i]) != get<1>(hotspots[j])  ){
				j++;
				hotspots[j] = hotspots[i];
			}
		}
		j++;
		hotspots.resize(j);
	}
	
	//INVERSE SORT BY HITS
//	rsort<true>(hotspots.data(), hotspots.data() + hotspots.size(), [] (const auto& v) { return get<2>(v); });
	sort(hotspots.begin(), hotspots.end(), 
		[] ( int_type_triple const& t1, int_type_triple const& t2) -> bool{ return get<2>(t1) > get<2>(t2); } );
	
};

template <uint8_t t_width, class int_type>
void fasta_read<t_width,int_type>::start_aligner(const string & EDS, const adjacency & adj,  uint32_t D, uint32_t max_a_c, uint32_t max_a_t, uint32_t max_a_o, fasta_read<t_width,int_type> & r_c){
	assert(filled_4);
	
	uint32_t aligns 	= 0;
	uint32_t tries	= 0;
	uint32_t i 	 	= 0;
	uint32_t i_r_c	= 0;

	//	map storing pair of position and error count
// 	std::map<int_type,uint32_t> pos_d; //TODO

	while(aligns < max_a_c	// only build max_a_c alignments
		&& tries < max_a_t	// only start aligner max_a_t times
		&& (i < hotspots.size() || i_r_c < r_c.hotspots.size() ) // another hint left
		)
	{		
		bool align_r_c = (i == hotspots.size()) // no further read in h_pos_i
					|| (i_r_c < r_c.hotspots.size()  && get<2>(hotspots[i]) < get<2>(r_c.hotspots[i_r_c])); // or next rev komp hint has more window hits 
		{
			tries++;
			tuple<string,uint32_t,int_type> levi_ali;
			if(align_r_c)
				levi_ali =  edsm_levinstein::levinstein<int_type>(EDS, adj, r_c.sequence, D, get<0>(r_c.hotspots[i_r_c]), get<1>(r_c.hotspots[i_r_c]));
			else
				levi_ali = edsm_levinstein::levinstein<int_type>(EDS, adj ,sequence, D, get<0>(hotspots[i]), get<1>(hotspots[i]) );
			
			if(get<2>(levi_ali) == EDS.size()){
				//no ali found
			}else{
// 				auto d = pos_d.find( get<2>(levi_ali) );//TODO
// 				if(d == pos_d.end() || d->second > get<1>(levi_ali)){// new pos or better//TODO alignment to existing
// 					pos_d[ get<2>(levi_ali) ] = get<1>(levi_ali);//TODO
					aligns++;
					alignment_type a = tie(get<0>(levi_ali),  get<1>(levi_ali), get<2>(levi_ali), align_r_c );
					alignments.push_back( a  );
// 				} 
				// else alignment already exists better or equal
			}
		}
		
		if(align_r_c)
			i_r_c++;
		else
			i++;
	}
	//rsort(alignments.data(), alignments.data() + alignments.size(),  [] (const auto& v) { return get<1>(v); });
	//SORT BY D
 	sort(alignments.begin(), alignments.end(),
 		[] ( alignment_type const& t1, alignment_type const& t2) -> bool{ return get<1>(t1) < get<1>(t2); } );
	
	
	
	uint32_t report_count = min(aligns, max_a_o);
	alignments.resize(report_count);
	delete_hints();
};
